import argparse

parser = argparse.ArgumentParser()
parser.add_argument('moves', nargs='+')
args = parser.parse_args()

while True:
    cmd_line = input()
    cmd = cmd_line.split(' ', 1)

    if cmd[0] == 'usi':
        print('id name test_usi_player')
        print('usiok')
    elif cmd[0] == 'setoption':
        pass
    elif cmd[0] == 'isready':
        print('readyok')
    elif cmd[0] == 'usinewgame':
        pass
    elif cmd[0] == 'position':
        moves = cmd[1].split(' ')[2:]
        if len(moves) < len(args.moves):
            bestmove = args.moves[len(moves)]
        else:
            bestmove = None
    elif cmd[0] == 'go':
        if bestmove:
            print('bestmove ' + bestmove)
        else:
            print('resign')
    elif cmd[0] == 'quit':
        break



import cshogi
from cshogi import Parser

COLOR_SYMBOLS = ['+', '-']
JAPANESE_END_GAMES = {
    '%TORYO': '投了',
    '%CHUDAN': '中断',
    '%SENNICHITE': '千日手',
    '%TIME_UP': '切れ負け',
    '%ILLEGAL_MOVE': '反則負け',
    '%JISHOGI': '持将棋',
    '%KACHI': '入力宣言',
}

class Exporter:
    def __init__(self, path=None, append=False, encoding=None):
        if path:
            self.open(path, append, encoding=encoding)
        else:
            self.f = None

    def open(self, path, append=False, encoding=None):
        self.f = open(path, 'a' if append else 'w', newline='\n', encoding=encoding)

    def close(self):
        self.f.close()

    def info(self, init_board=None, names=None, var_info=None, comment=None, version=None):
        if self.f.tell() != 0:
            self.f.write('/\n')
        if version:
                self.f.write(version)
                self.f.write('\n')
        if names:
            for name, turn in zip(names, ['+', '-']):
                self.f.write('N' + turn + name)
                self.f.write('\n')
        if comment:
            if comment[0] == "'":
                self.f.write(comment)
            else:
                self.f.write("'")
                self.f.write(comment)
                self.f.write('\n')
        if var_info:
            for k, v in var_info.items():
                self.f.write('$' + k + ':' + v)
                self.f.write('\n')
        if init_board:
            csa_pos = init_board.csa_pos()
            if csa_pos == 'P1-KY-KE-GI-KI-OU-KI-GI-KE-KY\nP2 * -HI *  *  *  *  * -KA * \nP3-FU-FU-FU-FU-FU-FU-FU-FU-FU\nP4 *  *  *  *  *  *  *  *  * \nP5 *  *  *  *  *  *  *  *  * \nP6 *  *  *  *  *  *  *  *  * \nP7+FU+FU+FU+FU+FU+FU+FU+FU+FU\nP8 * +KA *  *  *  *  * +HI * \nP9+KY+KE+GI+KI+OU+KI+GI+KE+KY\n+\n':
                self.f.write('PI\n+\n')
            else:
                self.f.write(csa_pos)
            self.turn = init_board.turn
        else:
            self.f.write('PI\n+\n')
            self.turn = cshogi.BLACK

    def move(self, move, time=None, comment=None, sep='\n'):
        self.f.write(COLOR_SYMBOLS[self.turn])
        self.f.write(cshogi.move_to_csa(move))
        if time is not None:
            self.f.write(sep)
            self.f.write('T' + str(time))
        if comment:
            self.f.write(sep)
            self.f.write("'" + comment)
        self.f.write('\n')
        self.turn = cshogi.opponent(self.turn)

    def endgame(self, endgame, time=None):
        self.f.write(endgame)
        self.f.write('\n')
        if time is not None:
            self.f.write('T' + str(time))
            self.f.write('\n')
        self.f.flush()



import cshogi
from datetime import datetime

PGN_SQUARE_NAMES = [
	'i9', 'i8', 'i7', 'i6', 'i5', 'i4', 'i3', 'i2', 'i1',
	'h9', 'h8', 'h7', 'h6', 'h5', 'h4', 'h3', 'h2', 'h1',
	'g9', 'g8', 'g7', 'g6', 'g5', 'g4', 'g3', 'g2', 'g1',
	'f9', 'f8', 'f7', 'f6', 'f5', 'f4', 'f3', 'f2', 'f1',
	'e9', 'e8', 'e7', 'e6', 'e5', 'e4', 'e3', 'e2', 'e1',
	'd9', 'd8', 'd7', 'd6', 'd5', 'd4', 'd3', 'd2', 'd1',
	'c9', 'c8', 'c7', 'c6', 'c5', 'c4', 'c3', 'c2', 'c1',
	'b9', 'b8', 'b7', 'b6', 'b5', 'b4', 'b3', 'b2', 'b1',
	'a9', 'a8', 'a7', 'a6', 'a5', 'a4', 'a3', 'a2', 'a1',
]

PGN_HAND_PIECES = [
    'P', 'L', 'N', 'S', 'G', 'B', 'R',
]

PGN_PIECE_TYPES = [
    None,
    'P', 'L', 'N', 'S', 'B', 'R', 'G', 'K',
    '+P', '+L', '+K', '+S', '+B', '+R',
]

def move_to_san(move):
    move_to = PGN_SQUARE_NAMES[cshogi.move_to(move)]

    if cshogi.move_is_drop(move):
        return PGN_HAND_PIECES[cshogi.move_drop_hand_piece(move)] + '@' + move_to

    move_from = PGN_SQUARE_NAMES[cshogi.move_from(move)]
    promotion = '+' if cshogi.move_is_promotion(move) else ''
    return PGN_PIECE_TYPES[cshogi.move_from_piece_type(move)] + move_from + move_to + promotion

class Exporter:
    def __init__(self, path=None, append=False):
        if path:
            self.open(path, append)
        else:
            self.f = None

    def open(self, path, append=False):
        self.f = open(path, 'a' if append else 'w', newline='\n')

    def close(self):
        self.f.close()

    def tag_pair(self, names, result, event='?', site='?', starttime=datetime.now(), round=1):
        self.f.write('[Event "' + event +'"]\n')
        self.f.write('[Site "' + site + '"]\n')
        self.f.write('[Date "' + (starttime.strftime('%Y.%m.%d') if starttime else '????.??.??') + '"]\n')
        self.f.write('[Round "' + str(round) + '"]\n')
        self.f.write('[White "' + names[0] + '"]\n')
        self.f.write('[Black "' + names[1] + '"]\n')
        if result == cshogi.BLACK_WIN:
            self.result_str = '1-0'
        elif result == cshogi.WHITE_WIN:
            self.result_str = '0-1'
        elif result == cshogi.DRAW:
            self.result_str = '1/2-1/2'
        else:
            self.result_str = '*'
        self.f.write('[Result "' + self.result_str + '"]\n\n')
        self.f.flush()

    def movetext(self, moves):
        line = ''
        for i, move in enumerate(moves):
            if i % 2 == 0:
                part = str(i // 2 + 1) + '. '
            else:
                part = ''
            part += move_to_san(moves[i])
            if i + 1 == len(moves):
                part += ' ' + self.result_str
            if len(line) + len(part) <= 80:
                if line != '':
                    line += ' '
                line += part
            else:
                self.f.write(line + '\n')
                line = part

        self.f.write(line + '\n\n')
        self.f.flush()



import random
import math
from time import perf_counter
import re
import os
import datetime
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor

from cshogi import *
from cshogi.usi import Engine
from cshogi import CSA
from cshogi import PGN
from cshogi.elo import Elo

try:
    is_jupyter = get_ipython().__class__.__name__ != 'TerminalInteractiveShell'
    if is_jupyter:
        from IPython.display import SVG, display
except NameError:
    is_jupyter = False

re_usi_info = re.compile('^.*score (cp|mate) ([+\-0-9]+).*pv (.*)$')
def to_score(m):
    if m[1] == 'cp':
        score = int(m[2])
    elif m[1] == 'mate':
        if m[2][0] == '-':
            score = -100000
        else:
            score = 100000
    return score

def usi_info_to_csa_comment(board, info):
    m = re_usi_info.match(info)
    if m is None:
        return None

    # score
    score = to_score(m) * (1 - board.turn * 2)

    # pv
    pv = []
    board2 = board.copy()
    for usi_move in m[3].split(' '):
        move = board2.move_from_usi(usi_move)
        if not board2.is_legal(move):
            break
        pv.append(CSA.COLOR_SYMBOLS[board2.turn] + move_to_csa(move))
        board2.push(move)

    if len(pv) > 1:
        return f"** {score} {' '.join(pv[1:])}"
    else:
        return f"** {score}"

def usi_info_to_score(info):
    m = re_usi_info.match(info)
    if m is None:
        return None

    return to_score(m)

def main(engine1, engine2, options1={}, options2={}, names=None, games=1, resign=None, mate_win=False,
         byoyomi=None, time=None, inc=None,
         draw=256, ponder=False, no_swap=False, opening=None, opening_moves=24, opening_seed=None, opening_index=None,
         keep_process=False,
         csa=None, multi_csa=False, pgn=None, no_pgn_moves=False, is_display=False, debug=False,
         print_summary=True, callback=None):
    engine1 = Engine(engine1, connect=False)
    engine2 = Engine(engine2, connect=False)

    # byoyomi
    if type(byoyomi) in (list, tuple):
        if len(byoyomi) >= 2:
            byoyomi1, byoyomi2 = byoyomi
        else:
            byoyomi1 = byoyomi2 = byoyomi[0]
    else:
        byoyomi1 = byoyomi2 = byoyomi

    # time
    if type(time) in (list, tuple):
        if len(time) >= 2:
            time1, time2 = time
        else:
            time1 = time2 = time[0]
    else:
        time1 = time2 = time

    # inc
    if type(inc) in (list, tuple):
        if len(inc) >= 2:
            inc1, inc2 = inc
        else:
            inc1 = inc2 = inc[0]
    else:
        inc1 = inc2 = inc

    # ponder
    if ponder:
        executor = ThreadPoolExecutor(max_workers=2)

    # debug
    if debug:
        class Listener:
            def __init__(self, id):
                self.id = id
                self.info = self.bestmove = ''

            def __call__(self, line):
                print(self.id + ':' + line)
                self.info = self.bestmove
                self.bestmove = line

        listener1 = Listener('1')
        listener2 = Listener('2')
    else:
        class Listener:
            def __init__(self):
                self.info = self.bestmove = ''

            def __call__(self, line):
                self.info = self.bestmove
                self.bestmove = line
        listener1 = listener2 = Listener()

    # CSA
    if csa and multi_csa:
        csa_exporter = CSA.Exporter(csa, append=True)

    # PGN
    if pgn:
        pgn_exporter = PGN.Exporter(pgn, append=True)

    # 初期局面読み込み
    if opening:
        opening_list = []
        with open(opening) as f:
            opening_list = [line.strip()[15:].split(' ') for line in f]
        # インデックス指定
        if opening_index is not None:
            opening_list = [opening_list[opening_index]]
        else:
            # シャッフル
            if opening_seed is not None:
                random.seed(opening_seed)
            random.shuffle(opening_list)

    board = Board()
    engine1_won = [0, 0, 0, 0, 0, 0]
    engine2_won = [0, 0, 0, 0, 0, 0]
    draw_count = 0
    WIN_DRAW = 2
    for n in range(games):
        # 先後入れ替え
        if no_swap or n % 2 == 0:
            engines_order = (engine1, engine2)
            options_order = (options1, options2)
            listeners_order = (listener1, listener2)
            byoyomi_order = (byoyomi1, byoyomi2)
            btime = time1
            wtime = time2
            binc = inc1
            winc = inc2
        else:
            engines_order = (engine2, engine1)
            options_order = (options2, options1)
            listeners_order = (listener2, listener1)
            byoyomi_order = (byoyomi2, byoyomi1)
            btime = time2
            wtime = time1
            binc = inc2
            winc = inc1

        # 接続とエンジン設定
        for engine, options, listener in zip(engines_order, options_order, listeners_order):
            if engine.proc is None:
                engine.connect(listener=listener)
            if ponder:
                engine.setoption('USI_Ponder', 'true', listener=listener)
            for name, value in options.items():
                engine.setoption(name, value, listener=listener)
            engine.isready(listener=listener)

        if names:
            if names[0]: engine1.name = names[0]
            if names[1]: engine2.name = names[1]

        print('{} vs {} start.'.format(engines_order[0].name, engines_order[1].name))

        # 初期局設定
        board.reset()
        moves = []
        usi_moves = []
        repetition_hash = defaultdict(int)
        if csa:
            engine_names = [engine.name for engine in engines_order]
            if not multi_csa:
                csa_exporter = CSA.Exporter(os.path.join(csa, '+'.join(engine_names) + '+' + datetime.datetime.now().strftime('%Y%m%d%H%M%S') + '.csa'), append=False)
            csa_exporter.info(board, engine_names, version='V2')
        if opening:
            for move_usi in opening_list[n // 2 % len(opening_list)]:
                move = board.push_usi(move_usi)
                if csa:
                    csa_exporter.move(move)
                moves.append(move)
                usi_moves.append(move_usi)
                repetition_hash[board.zobrist_hash()] += 1
                if board.move_number > opening_moves:
                    break

        # 盤面表示
        if is_display:
            print('開始局面')
            if is_jupyter:
                display(SVG(board.to_svg()))
            else:
                print(board)

        # 新規ゲーム
        for engine, listener in zip(engines_order, listeners_order):
            engine.usinewgame(listener=listener)

        # 対局
        is_game_over = False
        is_nyugyoku = False
        is_illegal = False
        is_repetition_win = False
        is_repetition_lose = False
        is_fourfold_repetition = False
        is_timeup = False
        remain_time = [btime, wtime]
        inc_time = (binc, winc)
        bestmove = None
        pondermove = None
        pre_pondermove = None
        feature = None
        pre_feature = None
        while not is_game_over:
            engine_index = (board.move_number - 1) % 2
            engine = engines_order[engine_index]
            listener = listeners_order[engine_index]
            byoyomi = byoyomi_order[engine_index]

            # 持将棋
            if board.move_number > draw:
                is_game_over = True
                break

            # ponder
            ponderhit = False
            if ponder:
                if pre_pondermove:
                    if bestmove == pre_pondermove:
                        ponderhit = True
                        pre_pondermove = pondermove

                        start_time = perf_counter()

                        engine.ponderhit(listener=listener)
                        bestmove, pondermove = pre_feature.result()
    
                        elapsed_time = perf_counter() - start_time
                    else:
                        pre_pondermove = pondermove
                        engine.stop(listener=listener)
                        pre_feature.result()
                else:
                    pre_pondermove = pondermove
                pre_feature = feature

            if not ponderhit:
                # position
                engine.position(usi_moves, listener=listener)

                start_time = perf_counter()

                # go
                bestmove, pondermove = engine.go(byoyomi=byoyomi, btime=remain_time[BLACK], wtime=remain_time[WHITE], binc=binc, winc=winc, listener=listener)

                elapsed_time = perf_counter() - start_time

            if remain_time[board.turn] is not None:
                if inc_time[board.turn] is not None:
                    remain_time[board.turn] += inc_time[board.turn]
                remain_time[board.turn] -= math.ceil(elapsed_time * 1000)

                if remain_time[board.turn] < 0:
                    # 1秒未満は切れ負けにしない
                    if remain_time[board.turn] > -1000:
                        remain_time[board.turn] = 0
                    else:
                        # 時間切れ負け
                        is_timeup = True
                        is_game_over = True
                        break

            # ponder
            if ponder and pondermove:
                engine.position(usi_moves + [bestmove, pondermove], listener=listener)
                feature = executor.submit(engine.go, ponder=True, byoyomi=byoyomi, btime=remain_time[BLACK], wtime=remain_time[WHITE], binc=binc, winc=winc, listener=listener)

            score = usi_info_to_score(listener.info)
            # 投了閾値
            if resign is not None:
                if score is not None and score <= -resign:
                    # 投了
                    is_game_over = True
                    break

            # 詰みを見つけたら終了
            if mate_win:
                if score is not None and score == 100000:
                    move = board.move_from_usi(bestmove)
                    if csa:
                        csa_exporter.move(move, time=int(elapsed_time), comment=usi_info_to_csa_comment(board, listener.info))
                    board.push(move)
                    is_game_over = True
                    break

            if bestmove == 'resign':
                # 投了
                is_game_over = True
                break
            elif bestmove == 'win':
                # 入玉勝ち宣言
                is_nyugyoku = True
                is_game_over = True
                break
            else:
                move = board.move_from_usi(bestmove)
                if board.is_legal(move):
                    if csa:
                        csa_exporter.move(move, time=int(elapsed_time), comment=usi_info_to_csa_comment(board, listener.info))
                    board.push(move)
                    moves.append(move)
                    usi_moves.append(bestmove)
                    key = board.zobrist_hash()
                    repetition_hash[key] += 1
                    # 千日手
                    if repetition_hash[key] == 4:
                        # 連続王手
                        is_draw = board.is_draw()
                        if is_draw == REPETITION_WIN:
                            is_repetition_win = True
                            is_game_over = True
                            break
                        elif is_draw == REPETITION_LOSE:
                            is_repetition_lose = True
                            is_game_over = True
                            break
                        is_fourfold_repetition = True
                        is_game_over = True
                        break
                else:
                    is_illegal = True
                    is_game_over = True
                    break

            # 盤面表示
            if is_display:
                print('{}手目'.format(len(usi_moves)))
                if is_jupyter:
                    display(SVG(board.to_svg(move)))
                else:
                    print(board)

            # 終局判定
            if board.is_game_over():
                is_game_over = True
                break

        # ゲーム終了
        for engine, listener in zip(engines_order, listeners_order):
            engine.gameover(listener=listener)

        # エンジン終了
        if not keep_process:
            for engine, listener in zip(engines_order, listeners_order):
                engine.quit(listener=listener)

        # 結果出力
        if not board.is_game_over() and board.move_number > draw:
            win = WIN_DRAW
            print('まで{}手で持将棋'.format(board.move_number - 1))
            csa_endgame = '%JISHOGI'
        elif is_fourfold_repetition:
            win = WIN_DRAW
            print('まで{}手で千日手'.format(board.move_number - 1))
            csa_endgame = '%SENNICHITE'
        elif is_nyugyoku:
            win = board.turn
            print('まで{}手で入玉宣言'.format(board.move_number - 1))
            csa_endgame = '%KACHI'
        elif is_illegal:
            win = opponent(board.turn)
            print('まで{}手で{}の反則負け'.format(board.move_number - 1, '先手' if win == WHITE else '後手'))
            csa_endgame = '%ILLEGAL_MOVE'
        elif is_repetition_win:
            win = board.turn
            print('まで{}手で{}の反則勝ち'.format(board.move_number - 1, '先手' if win == BLACK else '後手'))
            csa_endgame = '%+ILLEGAL_ACTION' if board.turn == WHITE else '%-ILLEGAL_ACTION'
        elif is_repetition_lose:
            win = opponent(board.turn)
            print('まで{}手で{}の反則負け'.format(board.move_number - 1, '先手' if win == WHITE else '後手'))
            csa_endgame = 'ILLEGAL_MOVE'
        elif is_timeup:
            win = opponent(board.turn)
            print('まで{}手で{}の切れ負け'.format(board.move_number - 1, '先手' if win == WHITE else '後手'))
            csa_endgame = '%TIME_UP'
        else:
            win = opponent(board.turn)
            print('まで{}手で{}の勝ち'.format(board.move_number - 1, '先手' if win == BLACK else '後手'))
            csa_endgame = '%TORYO'

        # 勝敗カウント
        if engines_order[0] is engine1:
            engine1_turn = 0
            engine2_turn = 1
        else:
            engine1_turn = 1
            engine2_turn = 0
        if win == WIN_DRAW:
            draw_count += 1
            engine1_won[4 + engine1_turn] += 1
            engine2_won[4 + engine2_turn] += 1
        elif engine1_turn == 0 and win == BLACK or engine1_turn == 1 and win == WHITE:
            engine1_won[engine1_turn] += 1
            engine2_won[2 + engine2_turn] += 1
        else:
            engine2_won[engine2_turn] += 1
            engine1_won[2 + engine1_turn] += 1

        black_won = engine1_won[0] + engine2_won[0]
        white_won = engine1_won[1] + engine2_won[1]
        engine1_won_sum = engine1_won[0] + engine1_won[1]
        engine2_won_sum = engine2_won[0] + engine2_won[1]
        total_count = engine1_won_sum + engine2_won_sum + draw_count

        # 勝敗状況表示
        if print_summary:
            print('{} of {} games finished.'.format(n + 1, games))
            print('{} vs {}: {}-{}-{} ({:.1f}%)'.format(
                engine1.name, engine2.name, engine1_won_sum, engine2_won_sum, draw_count,
                (engine1_won_sum + draw_count / 2) / total_count * 100))
            print('Black vs White: {}-{}-{} ({:.1f}%)'.format(
                black_won, white_won, draw_count,
                (black_won + draw_count / 2) / total_count * 100))
            print('{} playing Black: {}-{}-{} ({:.1f}%)'.format(
                engine1.name,
                engine1_won[0], engine1_won[2], engine1_won[4],
                (engine1_won[0] + engine1_won[4] / 2) / (engine1_won[0] + engine1_won[2] + engine1_won[4]) * 100))
            print('{} playing White: {}-{}-{} ({:.1f}%)'.format(
                engine1.name,
                engine1_won[1], engine1_won[3], engine1_won[5],
                (engine1_won[1] + engine1_won[5] / 2) / (engine1_won[1] + engine1_won[3] + engine1_won[5]) * 100 if not no_swap and n > 0 else 0))
            print('{} playing Black: {}-{}-{} ({:.1f}%)'.format(
                engine2.name,
                engine2_won[0], engine2_won[2], engine2_won[4],
                (engine2_won[0] + engine2_won[4] / 2) / (engine2_won[0] + engine2_won[2] + engine2_won[4]) * 100 if not no_swap and n > 0 else 0))
            print('{} playing White: {}-{}-{} ({:.1f}%)'.format(
                engine2.name,
                engine2_won[1], engine2_won[3], engine2_won[5],
                (engine2_won[1] + engine2_won[5] / 2) / (engine2_won[1] + engine2_won[3] + engine2_won[5]) * 100))
            elo = Elo(engine1_won_sum, engine2_won_sum, draw_count)
            if engine1_won_sum > 0 and engine2_won_sum > 0:
                try:
                    error_margin = elo.error_margin()
                except ValueError:
                    error_margin = math.nan
                print('Elo difference: {:.1f} +/- {:.1f}, LOS: {:.1f} %, DrawRatio: {:.1f} %'.format(
                    elo.diff(), error_margin, elo.los(), elo.draw_ratio()))

        # CSA
        if csa:
            csa_exporter.endgame(csa_endgame)

        # PGN
        if pgn:
            if win == BLACK:
                result = BLACK_WIN
            elif win == WHITE:
                result = WHITE_WIN
            else:
                result = DRAW
            pgn_exporter.tag_pair([engine.name for engine in engines_order], result, round=n+1)
            if not no_pgn_moves:
                pgn_exporter.movetext(moves)

        if callback:
            is_continue = callback({
                'engine1_name': engine1.name,
                'engine2_name': engine2.name,
                'engine1_won': engine1_won_sum,
                'engine2_won': engine2_won_sum,
                'black_won': black_won,
                'white_won': white_won,
                'draw': draw_count,
                'total': total_count,
                })
            if not is_continue:
                break

    # CSA
    if csa:
        csa_exporter.close()

    # PGN
    if pgn:
        pgn_exporter.close()

    # エンジン終了
    if keep_process:
        for engine, listener in zip(engines_order, listeners_order):
            engine.quit(listener=listener)

    return {
        'engine1_name': engine1.name,
        'engine2_name': engine2.name,
        'engine1_won': engine1_won_sum,
        'engine2_won': engine2_won_sum,
        'black_won': black_won,
        'white_won': white_won,
        'draw': draw_count,
        'total': total_count,
        }

if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('engine1')
    parser.add_argument('engine2')
    parser.add_argument('engine3', nargs='?')
    parser.add_argument('--options1', type=str, default='')
    parser.add_argument('--options2', type=str, default='')
    parser.add_argument('--options3', type=str, default='')
    parser.add_argument('--name1', type=str)
    parser.add_argument('--name2', type=str)
    parser.add_argument('--name3', type=str)
    parser.add_argument('--games', type=int, default=1)
    parser.add_argument('--resign', type=int)
    parser.add_argument('--mate-win', action='store_true')
    parser.add_argument('--byoyomi', type=int, nargs='+')
    parser.add_argument('--time', type=int, nargs='+')
    parser.add_argument('--inc', type=int, nargs='+')
    parser.add_argument('--draw', type=int, default=256)
    parser.add_argument('--ponder', action='store_true')
    parser.add_argument('--no-swap', action='store_true')
    parser.add_argument('--opening', type=str)
    parser.add_argument('--opening-moves', type=int, default=24)
    parser.add_argument('--opening-seed', type=int)
    parser.add_argument('--opening-index', type=int)
    parser.add_argument('--keep-process', action='store_true')
    parser.add_argument('--csa', type=str)
    parser.add_argument('--multi-csa', action='store_true')
    parser.add_argument('--pgn', type=str)
    parser.add_argument('--no-pgn-moves', action='store_true')
    parser.add_argument('--display', action='store_true')
    parser.add_argument('--debug', action='store_true')
    args = parser.parse_args()

    if args.csa is not None and not args.multi_csa:
        os.makedirs(args.csa, exist_ok=True)

    options_list = [{}, {}, {}]
    for i, kvs in enumerate([options.split(',') for options in (args.options1, args.options2, args.options3)]):
        if len(kvs) == 1 and kvs[0] == '':
            continue
        for kv_str in kvs:
            kv = kv_str.split(':', 1)
            if len(kv) != 2:
                raise ValueError('options{}'.format(i + 1))
            options_list[i][kv[0]] = kv[1]

    if args.engine3 is None:
        # 1 on 1 matches
        main(args.engine1, args.engine2,
            options_list[0], options_list[1],
            [args.name1, args.name2],
            args.games, args.resign, args.mate_win,
            args.byoyomi, args.time, args.inc,
            args.draw, args.ponder, args.no_swap, args.opening, args.opening_moves, args.opening_seed, args.opening_index,
            args.keep_process,
            args.csa, args.multi_csa,
            args.pgn, args.no_pgn_moves,
            args.display, args.debug)
    else:
        # league matches
        engines = (args.engine1, args.engine2, args.engine3)
        names = (args.name1, args.name2, args.name3)
        if args.byoyomi:
            if len(args.byoyomi) == 3:
                byoyomis = (args.byoyomi[0], args.byoyomi[1], args.byoyomi[2])
            else:
                byoyomis = (args.byoyomi[0], args.byoyomi[0], args.byoyomi[0])
        else:
            byoyomis = (None, None, None)
        if args.time:
            if len(args.time) == 3:
                times = (args.time[0], args.time[1], args.time[2])
            else:
                times = (args.time[0], args.time[0], args.time[0])
        else:
            times = (None, None, None)
        if args.inc:
            if len(args.inc) == 3:
                incs = (args.inc[0], args.inc[1], args.inc[2])
            else:
                incs = (args.inc[0], args.inc[0], args.inc[0])
        else:
            incs = (None, None, None)
        combinations = ((0, 1), (0, 2), (1, 2))
        results = [
            { 'engine1_won': 0, 'engine2_won': 0, 'draw': 0, 'total': 0 },
            { 'engine1_won': 0, 'engine2_won': 0, 'draw': 0, 'total': 0 },
            { 'engine1_won': 0, 'engine2_won': 0, 'draw': 0, 'total': 0 }]

        # 初期局面
        if args.opening:
            # インデックス指定
            if args.opening_index is not None:
                opening_index_list = [args.opening_index]
            else:
                with open(args.opening) as f:
                    opening_index_list = list(range(len(f.readlines())))
                # シャッフル
                if args.opening_seed is not None:
                    random.seed(args.opening_seed)
                random.shuffle(opening_index_list)

        for n in range(0, args.games, 2):
            if args.opening:
                opening_index = opening_index_list[n // 2 % len(opening_index_list)]
            else:
                opening_index = None
            for i, (a, b) in enumerate(combinations):
                # 先後入れ替えて1回ずつ対局
                result = main(engines[a], engines[b],
                    options_list[a], options_list[b],
                    [names[a], names[b]],
                    2, args.resign, args.mate_win,
                    (byoyomis[a], byoyomis[b]), (times[a], times[b]), (incs[a], incs[b]),
                    args.draw, args.ponder, False, args.opening, args.opening_moves, None, opening_index,
                    args.keep_process,
                    args.csa, args.multi_csa,
                    args.pgn, args.no_pgn_moves,
                    args.display, args.debug,
                    print_summary=False)
                results[i]['engine1_name'] = result['engine1_name']
                results[i]['engine2_name'] = result['engine2_name']
                results[i]['engine1_won'] += result['engine1_won']
                results[i]['engine2_won'] += result['engine2_won']
                results[i]['draw'] += result['draw']
                results[i]['total'] += result['total']

            # 勝敗状況表示
            print('{} of {} games finished.'.format(n + 2, args.games))
            for i, (a, b) in enumerate(combinations):
                print('{} vs {}: {}-{}-{} ({:.1f}%)'.format(
                    results[i]['engine1_name'], results[i]['engine2_name'],
                    results[i]['engine1_won'], results[i]['engine2_won'], results[i]['draw'],
                    (results[i]['engine1_won'] + results[i]['draw'] / 2) / results[i]['total'] * 100))



import sys
import math
from cshogi import Board, CSA, KIF, move_from_csa, BLACK, WHITE, opponent, REPETITION_WIN, REPETITION_LOSE
from cshogi.usi import Engine
from cshogi.cli import usi_info_to_score, usi_info_to_csa_comment, re_usi_info
from flask import Flask, render_template, Markup, request
from wsgiref.simple_server import make_server

TURN_SYMBOLS = ('▲', '△')
CSA_TURN_SYMBOLS = {'+': '▲', '-': '△'}

class Human:
    def __init__(self, human_input):
        self.board = Board()
        self.name = 'Human'
        self.human_input = human_input

    def usi(self, listener=None):
        return []

    def isready(self, listener=None):
        pass

    def usinewgame(self, listener=None):
        pass

    def position(self, moves=None, sfen="startpos", listener=None):
        self.board.reset()
        for move in moves:
            self.board.push_usi(move)

    def go(self, ponder=False, btime=None, wtime=None, byoyomi=None, binc=None, winc=None, nodes=None, listener=None):
        import time
        while True:
            human_input = dict(self.human_input)
            if human_input['number'] == self.board.move_number:
                usi_move = human_input['move']
                move = self.board.move_from_usi(usi_move)
                if self.board.is_legal(move):
                    return usi_move, None
            time.sleep(0.1)

    def quit(self, listener=None):
        pass

def usi_info_to_pv(board, info):
    m = re_usi_info.match(info)
    if m is None:
        return None

    # pv
    pv = []
    board2 = board.copy()
    for usi_move in m[3].split(' '):
        move = board2.move_from_usi(usi_move)
        if not board2.is_legal(move):
            break
        pv.append(TURN_SYMBOLS[board2.turn] + KIF.move_to_kif(move, board2.peek()))
        board2.push(move)

    return ' '.join(pv)

def match(moves, engine1=None, engine2=None, options1={}, options2={}, names=None, byoyomi=None, time=None, inc=None, draw=256, human_input=None, csa=None):
    from collections import defaultdict
    from time import perf_counter

    class Listener:
        def __init__(self):
            self.info = self.bestmove = ''

        def __call__(self, line):
            self.info = self.bestmove
            self.bestmove = line
    listener = Listener()

    # CSA
    if csa:
        csa_exporter = CSA.Exporter(csa, append=False)

    engines = []
    for engine in (engine1, engine2):
        if engine == 'human':
            engines.append(Human(human_input))
        else:
            engines.append(Engine(engine, connect=True))
    for engine, options in zip(engines, (options1, options2)):
        for name, value in options.items():
            engine.setoption(name, value, listener=listener)
        engine.isready(listener=listener)

    for i in range(2):
        if names[i]:
            engines[i].name = names[i]
        else:
            names[i] = engines[i].name

    board = Board()
    usi_moves = []
    repetition_hash = defaultdict(int)

    # 新規ゲーム
    for engine in engines:
        engine.usinewgame()

    if csa:
        csa_exporter.info(board, names, version='V2')

    # 対局
    is_game_over = False
    is_nyugyoku = False
    is_illegal = False
    is_repetition_win = False
    is_repetition_lose = False
    is_fourfold_repetition = False
    is_timeup = False
    remain_time = [time, time]
    while not is_game_over:
        engine_index = (board.move_number - 1) % 2
        engine = engines[engine_index]

        # 持将棋
        if board.move_number > draw:
            is_game_over = True
            break

        # position
        engine.position(usi_moves)

        start_time = perf_counter()

        # go
        listener.info = ''
        listener.bestmove = ''
        bestmove, _ = engine.go(byoyomi=byoyomi, btime=remain_time[BLACK], wtime=remain_time[WHITE], binc=inc, winc=inc, listener=listener)

        elapsed_time = perf_counter() - start_time

        if remain_time[board.turn] is not None:
            if inc_time[board.turn] is not None:
                remain_time[board.turn] += inc_time[board.turn]
            remain_time[board.turn] -= math.ceil(elapsed_time * 1000)

        if bestmove == 'resign':
            # 投了
            is_game_over = True
            break
        elif bestmove == 'win':
            # 入玉勝ち宣言
            is_nyugyoku = True
            is_game_over = True
            break
        else:
            move = board.move_from_usi(bestmove)
            if board.is_legal(move):
                if csa:
                    csa_exporter.move(move, time=int(elapsed_time), comment=usi_info_to_csa_comment(board, listener.info))
                score = usi_info_to_score(listener.info)
                pv = usi_info_to_pv(board, listener.info)
                moves.append({
                    'number': board.move_number,
                    'kif_move': TURN_SYMBOLS[(board.move_number + 1) % 2] + KIF.move_to_kif(move, board.peek()),
                    'time': math.ceil(elapsed_time),
                    'move': move,
                    'eval': (score * (1 if board.turn == BLACK else -1)) if score is not None else 'null',
                    'pv': pv if pv is not None else '',
                })
                board.push(move)
                usi_moves.append(bestmove)
                key = board.zobrist_hash()
                repetition_hash[key] += 1
                # 千日手
                if repetition_hash[key] == 4:
                    # 連続王手
                    is_draw = board.is_draw()
                    if is_draw == REPETITION_WIN:
                        is_repetition_win = True
                        is_game_over = True
                        break
                    elif is_draw == REPETITION_LOSE:
                        is_repetition_lose = True
                        is_game_over = True
                        break
                    is_fourfold_repetition = True
                    is_game_over = True
                    break
            else:
                is_illegal = True
                is_game_over = True
                break

        # 終局判定
        if board.is_game_over():
            is_game_over = True
            break

    # エンジン終了
    for engine in engines:
        engine.quit()

    # 結果出力
    if not board.is_game_over() and board.move_number > draw:
        result = '持将棋'
        csa_endgame = '%JISHOGI'
    elif is_fourfold_repetition:
        result = '千日手'
        csa_endgame = '%SENNICHITE'
    elif is_nyugyoku:
        result = TURN_SYMBOLS[board.turn] + '入玉宣言'
        csa_endgame = '%KACHI'
    elif is_illegal:
        win = opponent(board.turn)
        result = '{}の反則負け'.format('先手' if win == WHITE else '後手')
        csa_endgame = '%ILLEGAL_MOVE'
    elif is_repetition_win:
        win = board.turn
        result = '{}の反則勝ち'.format('先手' if win == BLACK else '後手')
        csa_endgame = '%+ILLEGAL_ACTION' if board.turn == WHITE else '%-ILLEGAL_ACTION'
    elif is_repetition_lose:
        win = opponent(board.turn)
        result = '{}の反則負け'.format('先手' if win == WHITE else '後手')
        csa_endgame = 'ILLEGAL_MOVE'
    elif is_timeup:
        win = opponent(board.turn)
        result = '{}の切れ負け'.format('先手' if win == WHITE else '後手')
        csa_endgame = '%TIME_UP'
    else:
        result = TURN_SYMBOLS[board.turn] + '投了'
        csa_endgame = '%TORYO'

    # CSA
    if csa:
        csa_exporter.endgame(csa_endgame)

    moves.append({
        'number': board.move_number,
        'kif_move': result,
        'time': 0,
        'move': 0,
        'eval': 'null',
        'pv': '',
    })

def run(engine1=None, engine2=None, options1={}, options2={}, name1=None, name2=None, byoyomi=None, time=None, inc=None, draw=256, csa=None, host='localhost', port=8000):
    is_match = 'false'
    auto_update = 'false'

    if engine1 and engine2:
        from multiprocessing import Process, Manager
        manager = Manager()
        moves = manager.list()
        human_input = manager.dict({ 'number': 0, 'move': None })
        names = manager.list([name1, name2])
        humans = [engine1 == 'human', engine2 == 'human']
        match_proc = Process(target=match, args=[moves, engine1, engine2, options1, options2, names, byoyomi, time, inc, draw, human_input, csa])
        match_proc.start()
        is_match = 'true'
        auto_update = 'true'
    elif csa:
        moves = []
        kif = CSA.Parser.parse_file(csa)[0]
        names = kif.names
        for i, (move, prev_move, time, comment) in enumerate(zip(kif.moves, [None] + kif.moves[:-1], kif.times, kif.comments)):
            comment_items = comment.split(' ')
            eval = 'null'
            pv = ''
            if len(comment_items) > 0 and comment_items[0] != '':
                eval = int(comment_items[0])
            if len(comment_items) > 1:
                for csa in comment_items[1:]:
                    pv += CSA_TURN_SYMBOLS[csa[0]] + KIF.move_to_kif(move_from_csa(csa[1:]))
            moves.append({
                'number': i + 1,
                'kif_move': TURN_SYMBOLS[i % 2] + KIF.move_to_kif(move, prev_move),
                'time': time,
                'move': move,
                'eval': eval,
                'pv': pv,
                })
        moves.append({
            'number': i + 2,
            'kif_move': TURN_SYMBOLS[(i + 1) % 2] + CSA.JAPANESE_END_GAMES[kif.endgame],
            'time': 0,
            'move': 0,
            'eval': 'null',
            'pv': '',
        })

    app = Flask(__name__)

    @app.route("/")
    def init_board():
        autoupdate = 'false'
        human = ''
        if is_match == 'true':
            if match_proc.is_alive():
                autoupdate = request.args.get('autoupdate', default=auto_update)
                if humans[len(moves) % 2]:
                    human = 'black' if len(moves) % 2 == 0 else 'white'
        return render_template('board.html', names=names, moves=moves, is_match=is_match, autoupdate=autoupdate, human=human)

    @app.route("/update")
    def update():
        human = ''
        if humans[len(moves) % 2]:
            human = 'black' if len(moves) % 2 == 0 else 'white'
        return { 'names': list(names), 'moves': list(moves), 'human':human }

    @app.route("/move")
    def human_move():
        import time
        human_input['move'] = request.args.get('move')
        human_input['number'] = int(request.args.get('number', default=0))
        time.sleep(0.2)
        human = ''
        if names[len(moves) % 2] == 'Human':
            human = 'black' if len(moves) % 2 == 0 else 'white'
        return { 'names': list(names), 'moves': list(moves), 'human':human }

    server = make_server(host, port, app)
    server.serve_forever()

def colab(engine1=None, engine2=None, options1={}, options2={}, name1=None, name2=None, byoyomi=None, time=None, inc=None, draw=256, csa=None):
    from multiprocessing import Process
    import portpicker
    from google.colab import output

    global proc
    if 'proc' in globals():
        proc.terminate()
        proc.join()

    port = portpicker.pick_unused_port()
    proc = Process(target=run, args=(engine1, engine2, options1, options2, name1, name2, byoyomi, time, inc, draw, csa, 'localhost', port))
    proc.start()
    output.serve_kernel_port_as_iframe(port, height='680')

if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('engine1')
    parser.add_argument('engine2')
    parser.add_argument('--options1', default='')
    parser.add_argument('--options2', default='')
    parser.add_argument('--name1')
    parser.add_argument('--name2')
    parser.add_argument('--byoyomi', type=int)
    parser.add_argument('--time', type=int)
    parser.add_argument('--inc', type=int)
    parser.add_argument('--draw', type=int, default=256)
    parser.add_argument('--csa')
    parser.add_argument('--host', type=str, default='localhost')
    parser.add_argument('--port', type=int, default=8000)
    args = parser.parse_args()

    options_list = [{}, {}]
    for i, kvs in enumerate([options.split(',') for options in (args.options1, args.options2)]):
        if len(kvs) == 1 and kvs[0] == '':
            continue
        for kv_str in kvs:
            kv = kv_str.split(':', 1)
            if len(kv) != 2:
                raise ValueError('options{}'.format(i + 1))
            options_list[i][kv[0]] = kv[1]

    run(engine1=args.engine1, engine2=args.engine2,
        options1=options_list[0], options2=options_list[1],
        name1=args.name1, name2=args.name2,
        byoyomi=args.byoyomi, time=args.time, inc=args.inc, draw=args.draw,
        csa=args.csa, host=args.host, port=args.port)



from cshogi._cshogi import _dlshogi_FEATURES1_NUM as FEATURES1_NUM
from cshogi._cshogi import _dlshogi_FEATURES2_NUM as FEATURES2_NUM
from cshogi._cshogi import _dlshogi_make_move_label as make_move_label

def make_input_features(board, features1, features2):
    board._dlshogi_make_input_features(features1, features2)



# This file is part of the python-shogi library.
# Copyright (C) 2015- Tasuku SUENAGA <tasuku-s-github@titech.ac>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import os
import re
import codecs
from datetime import datetime
import math

import cshogi

KIFU_TO_SQUARE_NAMES = [
    '１一', '１二', '１三', '１四', '１五', '１六', '１七', '１八', '１九',
    '２一', '２二', '２三', '２四', '２五', '２六', '２七', '２八', '２九',
    '３一', '３二', '３三', '３四', '３五', '３六', '３七', '３八', '３九',
    '４一', '４二', '４三', '４四', '４五', '４六', '４七', '４八', '４九',
    '５一', '５二', '５三', '５四', '５五', '５六', '５七', '５八', '５九',
    '６一', '６二', '６三', '６四', '６五', '６六', '６七', '６八', '６九',
    '７一', '７二', '７三', '７四', '７五', '７六', '７七', '７八', '７九',
    '８一', '８二', '８三', '８四', '８五', '８六', '８七', '８八', '８九',
    '９一', '９二', '９三', '９四', '９五', '９六', '９七', '９八', '９九',
]

KIFU_FROM_SQUARE_NAMES = [
    '11', '12', '13', '14', '15', '16', '17', '18', '19',
    '21', '22', '23', '24', '25', '26', '27', '28', '29',
    '31', '32', '33', '34', '35', '36', '37', '38', '39',
    '41', '42', '43', '44', '45', '46', '47', '48', '49',
    '51', '52', '53', '54', '55', '56', '57', '58', '59',
    '61', '62', '63', '64', '65', '66', '67', '68', '69',
    '71', '72', '73', '74', '75', '76', '77', '78', '79',
    '81', '82', '83', '84', '85', '86', '87', '88', '89',
    '91', '92', '93', '94', '95', '96', '97', '98', '99',
]

class ParserException(Exception):
    pass

class Parser:
    MOVE_RE = re.compile(r'\A *[0-9]+ (中断|投了|持将棋|先日手|詰み|切れ負け|反則勝ち|反則負け|(([１２３４５６７８９])([零一二三四五六七八九])|同　)([歩香桂銀金角飛玉と杏圭全馬龍])(打|(成?)\(([0-9])([0-9])\)))\s*(\([ /:0-9]+\))?\s*\Z')

    HANDYCAP_SFENS = {
        '平手': cshogi.STARTING_SFEN,
        '香落ち': 'lnsgkgsn1/1r5b1/ppppppppp/9/9/9/PPPPPPPPP/1B5R1/LNSGKGSNL w - 1',
        '右香落ち': '1nsgkgsnl/1r5b1/ppppppppp/9/9/9/PPPPPPPPP/1B5R1/LNSGKGSNL w - 1',
        '角落ち': 'lnsgkgsnl/1r7/ppppppppp/9/9/9/PPPPPPPPP/1B5R1/LNSGKGSNL w - 1',
        '飛車落ち': 'lnsgkgsnl/7b1/ppppppppp/9/9/9/PPPPPPPPP/1B5R1/LNSGKGSNL w - 1',
        '飛香落ち': 'lnsgkgsn1/7b1/ppppppppp/9/9/9/PPPPPPPPP/1B5R1/LNSGKGSNL w - 1',
        '二枚落ち': 'lnsgkgsnl/9/ppppppppp/9/9/9/PPPPPPPPP/1B5R1/LNSGKGSNL w - 1',
        '三枚落ち': 'lnsgkgsn1/9/ppppppppp/9/9/9/PPPPPPPPP/1B5R1/LNSGKGSNL w - 1',
        '四枚落ち': '1nsgkgsn1/9/ppppppppp/9/9/9/PPPPPPPPP/1B5R1/LNSGKGSNL w - 1',
        '五枚落ち': '2sgkgsn1/9/ppppppppp/9/9/9/PPPPPPPPP/1B5R1/LNSGKGSNL w - 1',
        '左五枚落ち': '1nsgkgs2/9/ppppppppp/9/9/9/PPPPPPPPP/1B5R1/LNSGKGSNL w - 1',
        '六枚落ち': '2sgkgs2/9/ppppppppp/9/9/9/PPPPPPPPP/1B5R1/LNSGKGSNL w - 1',
        '八枚落ち': '4k4/9/ppppppppp/9/9/9/PPPPPPPPP/1B5R1/LNSGKGSNL w - 1',
        'その他': None
    }

    RESULT_RE = re.compile(r'　*まで(\d+)手で((先|下|後|上)手の(勝ち|反則勝ち|反則負け)|千日手|持将棋|中断)')

    @staticmethod
    def parse_file(path):
        prefix, ext = os.path.splitext(path)
        enc = 'utf-8' if ext == '.kifu' else 'cp932'
        with codecs.open(path, 'r', enc) as f:
            return Parser.parse_str(f.read())

    @staticmethod
    def parse_pieces_in_hand(target):
        if target == 'なし': # None in japanese
            return {}

        result = {}
        for item in target.split('　'):
            if len(item) == 1:
                result[cshogi.PIECE_JAPANESE_SYMBOLS.index(item)] = 1
            elif len(item) == 2 or len(item) == 3:
                result[cshogi.PIECE_JAPANESE_SYMBOLS.index(item[0])] = \
                    cshogi.NUMBER_JAPANESE_KANJI_SYMBOLS.index(item[1:])
            elif len(item) == 0:
                pass
            else:
                raise ParserException('Invalid pieces in hand')
        return result

    @staticmethod
    def parse_move_str(line, last_to_square):
        # Normalize king/promoted kanji
        line = line.replace('王', '玉')
        line = line.replace('竜', '龍')
        line = line.replace('成銀', '全')
        line = line.replace('成桂', '圭')
        line = line.replace('成香', '杏')

        m = Parser.MOVE_RE.match(line)
        if m and m.group(1) not in [
                    '中断',
                    '投了',
                    '持将棋',
                    '千日手',
                    '詰み',
                    '切れ負け',
                    '反則勝ち',
                    '反則負け'
                ]:
            piece_type = cshogi.PIECE_JAPANESE_SYMBOLS.index(m.group(5))
            if m.group(2) == '同　':
                # same position
                to_square = last_to_square
            else:
                to_field = cshogi.NUMBER_JAPANESE_NUMBER_SYMBOLS.index(m.group(3)) - 1
                to_rank = cshogi.NUMBER_JAPANESE_KANJI_SYMBOLS.index(m.group(4)) - 1
                to_square = to_rank + to_field * 9
                last_to_square = to_square

            if m.group(6) == '打':
                # piece drop
                return ('{0}*{1}'.format(cshogi.PIECE_SYMBOLS[piece_type].upper(),
                    cshogi.SQUARE_NAMES[to_square]), last_to_square)
            else:
                from_field = int(m.group(8)) - 1
                from_rank = int(m.group(9)) - 1
                from_square = from_rank + from_field * 9

                promotion = (m.group(7) == '成')
                return (cshogi.SQUARE_NAMES[from_square] + cshogi.SQUARE_NAMES[to_square] + ('+' if promotion else ''), last_to_square)
        return (None, last_to_square)

    @staticmethod
    def parse_str(kif_str):
        line_no = 1

        starttime = None
        names = [None, None]
        pieces_in_hand = [{}, {}]
        current_turn = cshogi.BLACK
        sfen = cshogi.STARTING_SFEN
        moves = []
        last_to_square = None
        win = None
        kif_str = kif_str.replace('\r\n', '\n').replace('\r', '\n')
        for line in kif_str.split('\n'):
            if len(line) == 0 or line[0] == "*":
                pass
            elif '：' in line:
                (key, value) = line.split('：', 1)
                value = value.rstrip('　')
                if key == '開始日時':
                    try:
                        starttime = datetime.strptime(value, '%Y/%m/%d %H:%M:%S')
                    except ValueError:
                        try:
                            # if KIF file has not second information, try another parse
                            starttime = datetime.strptime(value, '%Y/%m/%d %H:%M') 
                        except ValueError:
                            pass

                if key == '先手' or key == '下手': # sente or shitate
                    # Blacks's name
                    names[cshogi.BLACK] = value
                elif key == '後手' or key == '上手': # gote or uwate
                    # White's name
                    names[cshogi.WHITE] = value
                elif key == '先手の持駒' or key == '下手の持駒': # sente or shitate's pieces in hand
                    # First player's pieces in hand
                    pieces_in_hand[cshogi.BLACK] == Parser.parse_pieces_in_hand(value)
                elif key == '後手の持駒' or key == '上手の持駒': # gote or uwate's pieces in hand
                    # Second player's pieces in hand
                    pieces_in_hand[cshogi.WHITE] == Parser.parse_pieces_in_hand(value)
                elif key == '手合割': # teai wari
                    sfen = Parser.HANDYCAP_SFENS[value]
                    if sfen is None:
                        raise ParserException('Cannot support handycap type "other"')
            elif line == '後手番':
                # Current turn is white
                current_turn = cshogi.WHITE
            else:
                (move, last_to_square) = Parser.parse_move_str(line, last_to_square)
                if move is not None:
                    moves.append(move)
                else:
                    m = Parser.RESULT_RE.match(line)
                    if m:
                        win_side_str = m.group(3)
                        if win_side_str == '先' or win_side_str == '下':
                            if m.group(4) == '反則負け':
                                win = 'w'
                            else:
                                win = 'b'
                        elif win_side_str == '後' or win_side_str == '上':
                            if m.group(4) == '反則負け':
                                win = 'b'
                            else:
                                win = 'w'
                        elif m.group(2) == '中断':
                            win = None
                        else:
                            # TODO: repetition of moves with continuous check
                            win = '-'
            line_no += 1

        summary = {
            'starttime': starttime,
            'names': names,
            'sfen': sfen,
            'moves': moves,
            'win': win
        }

        # NOTE: for the same interface with CSA parser
        return [summary]

def sec_to_time(sec):
    h, m_ = divmod(math.ceil(sec), 60*60)
    m, s = divmod(m_, 60)
    return h, m, s

def move_to_kif(move, prev_move=None):
    to_sq = cshogi.move_to(move)
    move_to = KIFU_TO_SQUARE_NAMES[to_sq]
    if prev_move:
        if cshogi.move_to(prev_move) == to_sq:
            move_to = "同　"
    if not cshogi.move_is_drop(move):
        from_sq = cshogi.move_from(move)
        move_piece = cshogi.PIECE_JAPANESE_SYMBOLS[cshogi.move_from_piece_type(move)]
        if cshogi.move_is_promotion(move):
            return '{}{}成({})'.format(
                move_to,
                move_piece,
                KIFU_FROM_SQUARE_NAMES[from_sq],
                )
        else:
            return '{}{}({})'.format(
                move_to,
                move_piece,
                KIFU_FROM_SQUARE_NAMES[from_sq],
                )
    else:
        move_piece = cshogi.HAND_PIECE_JAPANESE_SYMBOLS[cshogi.move_drop_hand_piece(move)]
        return '{}{}打'.format(
            move_to,
            move_piece
            )

class Exporter:
    def __init__(self, path=None):
        if path:
            self.open(path)
        else:
            self.kifu = None

    def open(self, path):
        _, ext = os.path.splitext(path)
        enc = 'utf-8' if ext == '.kifu' else 'cp932'
        self.kifu = open(path, 'w', encoding=enc)
        self.prev_move = None
        self.move_number = 1

    def close(self):
        self.kifu.close()

    def header(self, names, starttime=datetime.now()):
        self.kifu.write('開始日時：' + starttime.strftime('%Y/%m/%d %H:%M:%S\n'))
        self.kifu.write('手合割：平手\n')
        self.kifu.write('先手：' + names[0] + '\n')
        self.kifu.write('後手：' + names[1] + '\n')
        self.kifu.write('手数----指手---------消費時間--\n')

    def move(self, move, sec=0, sec_sum=0):
        m, s = divmod(math.ceil(sec), 60)
        h_sum, m_sum, s_sum = sec_to_time(sec_sum)

        if cshogi.move_is_drop(move):
            padding = '    '
        elif cshogi.move_is_promotion(move):
            padding = ''
        else:
            padding = '  '
        move_str = move_to_kif(move, self.prev_move) + padding

        self.kifu.write('{:>4} {}      ({:>2}:{:02}/{:02}:{:02}:{:02})\n'.format(
            self.move_number,
            move_str,
            m, s,
            h_sum, m_sum, s_sum))

        self.move_number += 1
        self.prev_move = move

    def end(self, reason, sec=0, sec_sum=0):
        m, s = divmod(math.ceil(sec), 60)
        h_sum, m_sum, s_sum = sec_to_time(sec_sum)

        if reason == 'resign':
            move_str = '投了        '
        elif reason == 'win':
            move_str = '入玉宣言    '
        elif reason == 'draw':
            move_str = '持将棋      '
        elif reason == 'sennichite':
            move_str = '千日手      '
        elif reason == 'illegal_win':
            move_str = '反則勝ち    '
        elif reason == 'illegal_lose':
            move_str = '反則負け    '

        self.kifu.write('{:>4} {}      ({:>2}:{:02}/{:02}:{:02}:{:02})\n'.format(
            self.move_number,
            move_str,
            m, s,
            h_sum, m_sum, s_sum))

        # 結果出力
        if reason == 'resign':
            self.kifu.write('まで{}手で{}の勝ち\n'.format(self.move_number - 1, '先手' if self.move_number % 2 == 0 else '後手'))
        elif reason == 'draw':
            self.kifu.write('まで{}手で持将棋\n'.format(self.move_number + 1))
        elif reason == 'win':
            print('まで{}手で入玉宣言\n'.format(self.move_number - 1))
        elif reason == 'sennichite':
            self.kifu.write('まで{}手で千日手\n'.format(self.move_number - 1))
        elif reason == 'illegal_win':
            self.kifu.write('まで{}手で{}の反則勝ち\n'.format(self.move_number - 1, '先手' if self.move_number % 2 == 0 else '後手'))
        elif reason == 'illegal_lose':
            self.kifu.write('まで{}手で{}の反則負け\n'.format(self.move_number - 1, '先手' if self.move_number % 2 == 0 else '後手'))

    def info(self, info):
        turn = self.move_number % 2
        items = info.split(' ')
        comment = '**対局'
        i = 1
        while i < len(items):
            if items[i] == 'time':
                i += 1
                m, s = divmod(int(items[i]) / 1000, 60)
                s_str = '{:.1f}'.format(s)
                if s_str[1:2] == '.':
                    s_str = '0' + s_str
                comment += ' 時間 {:>02}:{}'.format(int(m), s_str)
            elif items[i] == 'depth':
                i += 1
                comment += ' 深さ {}'.format(items[i])
            elif items[i] == 'nodes':
                i += 1
                comment += ' ノード数 {}'.format(items[i])
            elif items[i] == 'score':
                i += 1
                if items[i] == 'cp':
                    i += 1
                    comment += ' 評価値 {}'.format(items[i] if turn == cshogi.BLACK else -int(items[i]))
                elif items[i] == 'mate':
                    i += 1
                    if items[i][0:1] == '+':
                        comment += ' +詰' if turn == cshogi.BLACK else ' -詰'
                    else:
                        comment += ' -詰' if turn == cshogi.BLACK else ' +詰'
                    comment += str(items[i][1:])
            else:
                i += 1
        self.kifu.write(comment + '\n')



import gym
from gym import spaces
import numpy as np
import cshogi
from collections import defaultdict

class ShogiEnv(gym.Env):
	metadata = {'render.modes': ['human', 'svg', 'ansi', 'sfen']}

	def __init__(self):
		cdef unsigned long long key

		super().__init__()

		self.board = cshogi.Board()
		self.moves = []
		self.repetition_hash = defaultdict(int)
		key = self.board.zobrist_hash()
		self.repetition_hash[key] += 1
		self.is_draw = None

		self.observation_space = spaces.Box(0, len(cshogi.PIECES)-1, (9, 9), dtype=np.uint8)

		# actionはmoveを直接受け付ける
		# sample()は非合法手も含む
		self.action_space = gym.spaces.Discrete(16777216)

	def reset(self, sfen=None, hcp=None):
		cdef unsigned long long key

		if not sfen is None:
			self.board.set_sfen(sfen)
		elif not hcp is None:
			self.board.set_hcp(hcp)
		else:
			self.board.reset()

		self.moves.clear()
		self.repetition_hash.clear()
		key = self.board.zobrist_hash()
		self.repetition_hash[key] += 1
		self.is_draw = None

		return self.board

	def render(self, str mode='human'):
		if mode == 'svg':
			if len(self.moves) > 0:
				return self.board.to_svg(lastmove=self.moves[-1])
			else:
				return self.board.to_svg()
		elif mode == 'ansi':
			print(self.board)
		elif mode == 'sfen':
			print(self.board.sfen())
		else:
			return self.board

	def step(self, int move):
		cdef unsigned long long key

		# 投了
		if move == 0:
			reward = -1.0
			done = True
			return reward, done, None

		assert self.board.is_legal(move)

		self.board.push(move)
		self.moves.append(move)

		key = self.board.zobrist_hash()
		self.repetition_hash[key] += 1
		# 千日手
		if self.repetition_hash[key] == 4:
			done = True
			# 連続王手
			self.is_draw = self.board.is_draw()
			if self.is_draw == cshogi.REPETITION_WIN:
				# 相手の手番なので報酬は反対になる
				reward = -1.0
			elif self.is_draw == cshogi.REPETITION_LOSE:
				reward = 1.0
			else:
				reward = 0.0
		else:
			done = self.board.is_game_over()
			if done:
				reward = 1.0
			else:
				reward = 0.0

		# Observationは直接フィールドを参照する
		return reward, done, self.is_draw



import math

def elo_diff(p):
	return -400.0 * math.log10(1.0 / p - 1.0)

def erf_inv(x):
	pi = 3.1415926535897

	a = 8.0 * (pi - 3.0) / (3.0 * pi * (4.0 - pi))
	y = math.log(1.0 - x * x)
	z = 2.0 / (pi * a) + y / 2.0

	ret = math.sqrt(math.sqrt(z * z - y / a) - z)

	if x < 0.0:
		return -ret
	return ret

def phi_inv(p):
	return math.sqrt(2.0) * erf_inv(2.0 * p - 1.0)

class Elo:
	def __init__(self, wins, losses, draws):
		self.wins = wins
		self.losses = losses
		self.draws = draws

		n = wins + losses + draws
		w = wins / n
		l = losses / n
		d = draws / n
		self.mu = w + d / 2.0

		dev_w = w * math.pow(1.0 - self.mu, 2.0)
		dev_l = l * math.pow(0.0 - self.mu, 2.0)
		dev_d = d * math.pow(0.5 - self.mu, 2.0)
		self.stdev = math.sqrt(dev_w + dev_l + dev_d) / math.sqrt(n)

	def point_ratio(self):
		total = (self.wins + self.losses + self.draws) * 2
		return ((self.wins * 2) + self.draws) / total

	def draw_ratio(self):
		n = self.wins + self.losses + self.draws
		return self.draws / n

	def diff(self):
		return elo_diff(self.mu)

	def error_margin(self):
		mu_min = self.mu + phi_inv(0.025) * self.stdev
		mu_max = self.mu + phi_inv(0.975) * self.stdev
		return (elo_diff(mu_max) - elo_diff(mu_min)) / 2.0

	# likelihood of superiority
	def los(self):
		return 100 * (0.5 + 0.5 * math.erf((self.wins - self.losses) / math.sqrt(2.0 * (self.wins + self.losses))))



from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext

class my_build_ext(build_ext):
    def build_extensions(self):
        if self.compiler.compiler_type == 'unix':
            for e in self.extensions:
                e.extra_compile_args = ['-msse4.2', '-mavx2']

        build_ext.build_extensions(self)

    def finalize_options(self):
        build_ext.finalize_options(self)
        __builtins__.__NUMPY_SETUP__ = False
        import numpy
        self.include_dirs.append(numpy.get_include())

ext_modules = [
    Extension('cshogi._cshogi',
        ['cshogi/_cshogi.pyx',
         "src/bitboard.cpp", "src/common.cpp", "src/generateMoves.cpp", "src/hand.cpp", "src/init.cpp", "src/move.cpp", "src/mt64bit.cpp", "src/position.cpp", "src/search.cpp", "src/square.cpp", "src/usi.cpp", "src/book.cpp", "src/mate.cpp", "src/dfpn.cpp"],
        language='c++',
        include_dirs = ["src"],
        define_macros=[('HAVE_SSE4', None), ('HAVE_SSE42', None), ('HAVE_AVX2', None)]),
    Extension('cshogi.gym_shogi.envs.shogi_env',
        ['cshogi/gym_shogi/envs/shogi_env.pyx'],
        language='c++'),
    Extension('cshogi.gym_shogi.envs.shogi_vec_env',
        ['cshogi/gym_shogi/envs/shogi_vec_env.pyx'],
        language='c++'),
]

setup(
    name='cshogi',
    version='0.4.1',
    packages=['cshogi', 'cshogi.usi', 'cshogi.gym_shogi', 'cshogi.gym_shogi.envs', 'cshogi.dlshogi', 'cshogi.web', 'cshogi.web.templates', 'cshogi.web.static'],
    package_data={'cshogi.web.templates': ['*'], 'cshogi.web.static': ['*']},
    ext_modules=ext_modules,
    cmdclass={'build_ext': my_build_ext},
    author='Tadao Yamaoka',
    url='https://github.com/TadaoYamaoka/cshogi',
    description = 'A fast Python shogi library',
    classifiers=[
        "Programming Language :: Python :: 3",
        'License :: OSI Approved :: GNU General Public License (GPL)',
        "Operating System :: OS Independent",
    ],
)


from libcpp.string cimport string
from libcpp.vector cimport vector
from libcpp cimport bool

import numpy as np
cimport numpy as np

import locale

dtypeHcp = np.dtype((np.uint8, 32))
dtypeEval = np.dtype(np.int16)
dtypeMove16 = np.dtype(np.int16)
dtypeGameResult = np.dtype(np.int8)

HuffmanCodedPos = np.dtype([
	('hcp', dtypeHcp),
	])

HuffmanCodedPosAndEval = np.dtype([
	('hcp', dtypeHcp),
	('eval', dtypeEval),
	('bestMove16', dtypeMove16),
	('gameResult', dtypeGameResult),
	('dummy', np.uint8),
	])

PackedSfen = np.dtype([
	('sfen', np.uint8, 32),
	])

PackedSfenValue = np.dtype([
	('sfen', np.uint8, 32),
	('score', np.int16),
	('move', np.uint16),
	('gamePly', np.uint16),
	('game_result', np.int8),
	('padding', np.uint8),
	])

dtypeKey = np.dtype(np.uint64)
BookEntry = np.dtype([
	('key', dtypeKey),
	('fromToPro', dtypeMove16),
	('count', np.uint16),
	('score', np.int32),
	])


STARTING_SFEN = 'lnsgkgsnl/1r5b1/ppppppppp/9/9/9/PPPPPPPPP/1B5R1/LNSGKGSNL b - 1'

SQUARES = [
	A1, B1, C1, D1, E1, F1, G1, H1, I1,
	A2, B2, C2, D2, E2, F2, G2, H2, I2,
	A3, B3, C3, D3, E3, F3, G3, H3, I3,
	A4, B4, C4, D4, E4, F4, G4, H4, I4,
	A5, B5, C5, D5, E5, F5, G5, H5, I5,
	A6, B6, C6, D6, E6, F6, G6, H6, I6,
	A7, B7, C7, D7, E7, F7, G7, H7, I7,
	A8, B8, C8, D8, E8, F8, G8, H8, I8,
	A9, B9, C9, D9, E9, F9, G9, H9, I9,
] = range(81)

SQUARE_NAMES = [
	'1a', '1b', '1c', '1d', '1e', '1f', '1g', '1h', '1i',
	'2a', '2b', '2c', '2d', '2e', '2f', '2g', '2h', '2i',
	'3a', '3b', '3c', '3d', '3e', '3f', '3g', '3h', '3i',
	'4a', '4b', '4c', '4d', '4e', '4f', '4g', '4h', '4i',
	'5a', '5b', '5c', '5d', '5e', '5f', '5g', '5h', '5i',
	'6a', '6b', '6c', '6d', '6e', '6f', '6g', '6h', '6i',
	'7a', '7b', '7c', '7d', '7e', '7f', '7g', '7h', '7i',
	'8a', '8b', '8c', '8d', '8e', '8f', '8g', '8h', '8i',
	'9a', '9b', '9c', '9d', '9e', '9f', '9g', '9h', '9i',
]

COLORS = [BLACK, WHITE] = range(2)

GAME_RESULTS = [
	DRAW, BLACK_WIN, WHITE_WIN,
] = range(3)

PIECE_TYPES_WITH_NONE = [NONE,
		   PAWN,      LANCE,      KNIGHT,      SILVER,
		 BISHOP,       ROOK,
		   GOLD,
		   KING,
	  PROM_PAWN, PROM_LANCE, PROM_KNIGHT, PROM_SILVER,
	PROM_BISHOP,  PROM_ROOK,
] = range(15)

PIECE_TYPES = [
		   PAWN,      LANCE,      KNIGHT,      SILVER,
		 BISHOP,       ROOK,
		   GOLD,
		   KING,
	  PROM_PAWN, PROM_LANCE, PROM_KNIGHT, PROM_SILVER,
	PROM_BISHOP,  PROM_ROOK,
]

PIECES = [NONE,
		  BPAWN,      BLANCE,      BKNIGHT,      BSILVER,
		BBISHOP,       BROOK,
		  BGOLD,
		  BKING,
	 BPROM_PAWN, BPROM_LANCE, BPROM_KNIGHT, BPROM_SILVER,
   BPROM_BISHOP,  BPROM_ROOK,       NOTUSE,       NOTUSE,
		  WPAWN,      WLANCE,      WKNIGHT,      WSILVER,
		WBISHOP,       WROOK,
		  WGOLD,
		  WKING,
	 WPROM_PAWN, WPROM_LANCE, WPROM_KNIGHT, WPROM_SILVER,
   WPROM_BISHOP, WPROM_ROOK,
] = range(31)

HAND_PIECES = [
		  HPAWN,     HLANCE,     HKNIGHT,     HSILVER,
		  HGOLD,
		HBISHOP,      HROOK,
] = range(7)

MAX_PIECES_IN_HAND = [
	18, 4, 4, 4,
	4,
	2, 2,
]

MOVE_NONE = 0

REPETITION_TYPES = [
	NOT_REPETITION, REPETITION_DRAW, REPETITION_WIN, REPETITION_LOSE,
	REPETITION_SUPERIOR, REPETITION_INFERIOR
] = range(6)

SVG_PIECE_DEFS = [
	'<g id="black-pawn"><text font-family="serif" font-size="17" text-anchor="middle" x="10.5" y="16.5">歩</text></g>',
	'<g id="black-lance"><text font-family="serif" font-size="17" text-anchor="middle" x="10.5" y="16.5">香</text></g>',
	'<g id="black-knight"><text font-family="serif" font-size="17" text-anchor="middle" x="10.5" y="16.5">桂</text></g>',
	'<g id="black-silver"><text font-family="serif" font-size="17" text-anchor="middle" x="10.5" y="16.5">銀</text></g>',
	'<g id="black-gold"><text font-family="serif" font-size="17" text-anchor="middle" x="10.5" y="16.5">金</text></g>',
	'<g id="black-bishop"><text font-family="serif" font-size="17" text-anchor="middle" x="10.5" y="16.5">角</text></g>',
	'<g id="black-rook"><text font-family="serif" font-size="17" text-anchor="middle" x="10.5" y="16.5">飛</text></g>',
	'<g id="black-king"><text font-family="serif" font-size="17" text-anchor="middle" x="10.5" y="16.5">王</text></g>',
	'<g id="black-pro-pawn"><text font-family="serif" font-size="17" text-anchor="middle" x="10.5" y="16.5">と</text></g>',
	'<g id="black-pro-lance" transform="scale(1.0, 0.5)"><text font-family="serif" font-size="17" text-anchor="middle" x="10.5" y="18">成</text><text font-family="serif" font-size="17" text-anchor="middle" x="10.5" y="34">香</text></g>',
	'<g id="black-pro-knight" transform="scale(1.0, 0.5)"><text font-family="serif" font-size="17" text-anchor="middle" x="10.5" y="18">成</text><text font-family="serif" font-size="17" text-anchor="middle" x="10.5" y="34">桂</text></g>',
	'<g id="black-pro-silver" transform="scale(1.0, 0.5)"><text font-family="serif" font-size="17" text-anchor="middle" x="10.5" y="18">成</text><text font-family="serif" font-size="17" text-anchor="middle" x="10.5" y="34">銀</text></g>',
	'<g id="black-horse"><text font-family="serif" font-size="17" text-anchor="middle" x="10.5" y="16.5">馬</text></g>',
	'<g id="black-dragon"><text font-family="serif" font-size="17" text-anchor="middle" x="10.5" y="16.5">龍</text></g>',
	'<g id="white-pawn" transform="rotate(180)"><text font-family="serif" font-size="17" text-anchor="middle" x="-10.5" y="-3.5">歩</text></g>',
	'<g id="white-lance" transform="rotate(180)"><text font-family="serif" font-size="17" text-anchor="middle" x="-10.5" y="-3.5">香</text></g>',
	'<g id="white-knight" transform="rotate(180)"><text font-family="serif" font-size="17" text-anchor="middle" x="-10.5" y="-3.5">桂</text></g>',
	'<g id="white-silver" transform="rotate(180)"><text font-family="serif" font-size="17" text-anchor="middle" x="-10.5" y="-3.5">銀</text></g>',
	'<g id="white-gold" transform="rotate(180)"><text font-family="serif" font-size="17" text-anchor="middle" x="-10.5" y="-3.5">金</text></g>',
	'<g id="white-bishop" transform="rotate(180)"><text font-family="serif" font-size="17" text-anchor="middle" x="-10.5" y="-3.5">角</text></g>',
	'<g id="white-rook" transform="rotate(180)"><text font-family="serif" font-size="17" text-anchor="middle" x="-10.5" y="-3.5">飛</text></g>',
	'<g id="white-king" transform="rotate(180)"><text font-family="serif" font-size="17" text-anchor="middle" x="-10.5" y="-3.5">王</text></g>',
	'<g id="white-pro-pawn" transform="rotate(180)"><text font-family="serif" font-size="17" text-anchor="middle" x="-10.5" y="-3.5">と</text></g>',
	'<g id="white-pro-lance" transform="scale(1.0, 0.5) rotate(180)"><text font-family="serif" font-size="17" text-anchor="middle" x="-10.5" y="-22">成</text><text font-family="serif" font-size="17" text-anchor="middle" x="-10.5" y="-6">香</text></g>',
	'<g id="white-pro-knight" transform="scale(1.0, 0.5) rotate(180)"><text font-family="serif" font-size="17" text-anchor="middle" x="-10.5" y="-22">成</text><text font-family="serif" font-size="17" text-anchor="middle" x="-10.5" y="-6">桂</text></g>',
	'<g id="white-pro-silver" transform="scale(1.0, 0.5) rotate(180)"><text font-family="serif" font-size="17" text-anchor="middle" x="-10.5" y="-22">成</text><text font-family="serif" font-size="17" text-anchor="middle" x="-10.5" y="-6">銀</text></g>',
	'<g id="white-horse" transform="rotate(180)"><text font-family="serif" font-size="17" text-anchor="middle" x="-10.5" y="-3.5">馬</text></g>',
	'<g id="white-dragon" transform="rotate(180)"><text font-family="serif" font-size="17" text-anchor="middle" x="-10.5" y="-3.5">龍</text></g>',
]
SVG_PIECE_DEF_IDS = [None,
	"black-pawn", "black-lance", "black-knight", "black-silver",
	"black-bishop", "black-rook",
	"black-gold",
	"black-king",
	"black-pro-pawn", "black-pro-lance", "black-pro-knight", "black-pro-silver",
	"black-horse", "black-dragon", None, None,
	"white-pawn", "white-lance", "white-knight", "white-silver",
	"white-bishop", "white-rook",
	"white-gold",
	"white-king",
	"white-pro-pawn", "white-pro-lance", "white-pro-knight", "white-pro-silver",
	"white-horse", "white-dragon",
]
NUMBER_JAPANESE_NUMBER_SYMBOLS = [ None, '１', '２', '３', '４', '５', '６', '７', '８', '９' ]
NUMBER_JAPANESE_KANJI_SYMBOLS = [ None, "一", "二", "三", "四", "五", "六", "七", "八", "九", "十" ]
SVG_SQUARES = '<g stroke="black"><rect x="20" y="10" width="181" height="181" fill="none" stroke-width="1.5" /><line x1="20.5" y1="30.5" x2="200.5" y2="30.5" stroke-width="1.0" /><line x1="20.5" y1="50.5" x2="200.5" y2="50.5" stroke-width="1.0" /><line x1="20.5" y1="70.5" x2="200.5" y2="70.5" stroke-width="1.0" /><line x1="20.5" y1="90.5" x2="200.5" y2="90.5" stroke-width="1.0" /><line x1="20.5" y1="110.5" x2="200.5" y2="110.5" stroke-width="1.0" /><line x1="20.5" y1="130.5" x2="200.5" y2="130.5" stroke-width="1.0" /><line x1="20.5" y1="150.5" x2="200.5" y2="150.5" stroke-width="1.0" /><line x1="20.5" y1="170.5" x2="200.5" y2="170.5" stroke-width="1.0" /><line x1="40.5" y1="10.5" x2="40.5" y2="190.5" stroke-width="1.0" /><line x1="60.5" y1="10.5" x2="60.5" y2="190.5" stroke-width="1.0" /><line x1="80.5" y1="10.5" x2="80.5" y2="190.5" stroke-width="1.0" /><line x1="100.5" y1="10.5" x2="100.5" y2="190.5" stroke-width="1.0" /><line x1="120.5" y1="10.5" x2="120.5" y2="190.5" stroke-width="1.0" /><line x1="140.5" y1="10.5" x2="140.5" y2="190.5" stroke-width="1.0" /><line x1="160.5" y1="10.5" x2="160.5" y2="190.5" stroke-width="1.0" /><line x1="180.5" y1="10.5" x2="180.5" y2="190.5" stroke-width="1.0" /></g>'
SVG_COORDINATES = '<g><text font-family="serif" text-anchor="middle" font-size="9" x="30.5" y="8">9</text><text font-family="serif" text-anchor="middle" font-size="9" x="50.5" y="8">8</text><text font-family="serif" text-anchor="middle" font-size="9" x="70.5" y="8">7</text><text font-family="serif" text-anchor="middle" font-size="9" x="90.5" y="8">6</text><text font-family="serif" text-anchor="middle" font-size="9" x="110.5" y="8">5</text><text font-family="serif" text-anchor="middle" font-size="9" x="130.5" y="8">4</text><text font-family="serif" text-anchor="middle" font-size="9" x="150.5" y="8">3</text><text font-family="serif" text-anchor="middle" font-size="9" x="170.5" y="8">2</text><text font-family="serif" text-anchor="middle" font-size="9" x="190.5" y="8">1</text><text font-family="serif" font-size="9" x="203.5" y="23">一</text><text font-family="serif" font-size="9" x="203.5" y="43">二</text><text font-family="serif" font-size="9" x="203.5" y="63">三</text><text font-family="serif" font-size="9" x="203.5" y="83">四</text><text font-family="serif" font-size="9" x="203.5" y="103">五</text><text font-family="serif" font-size="9" x="203.5" y="123">六</text><text font-family="serif" font-size="9" x="203.5" y="143">七</text><text font-family="serif" font-size="9" x="203.5" y="163">八</text><text font-family="serif" font-size="9" x="203.5" y="183">九</text></g>'
PIECE_SYMBOLS = [
	'',
	'p', 'l', 'n', 's', 'b', 'r', 'g', 'k', '+p', '+l', '+n', '+s', '+b', '+r'
]
PIECE_JAPANESE_SYMBOLS = [
	'',
	'歩', '香', '桂', '銀', '角', '飛', '金', '玉', 'と', '杏', '圭', '全', '馬', '龍'
]
HAND_PIECE_JAPANESE_SYMBOLS = [
	"歩", "香", "桂", "銀",
	"金",
	"角", "飛"
]

class SvgWrapper(str):
	def _repr_svg_(self):
		return self

cdef extern from "init.hpp":
	void initTable()

initTable()

cdef extern from "position.hpp":
	cdef cppclass Position:
		@staticmethod
		void initZobrist()

Position.initZobrist()

cdef extern from "cshogi.h":
	void HuffmanCodedPos_init()
	void PackedSfen_init()
	void Book_init()

HuffmanCodedPos_init()
PackedSfen_init()
Book_init()

cdef extern from "cshogi.h":
	string __to_usi(const int move)
	string __to_csa(const int move)

def to_usi(int move):
	return __to_usi(move)

def to_csa(int move):
	return __to_csa(move)

cdef extern from "cshogi.h":
	cdef cppclass __Board:
		__Board() except +
		__Board(const string& sfen) except +
		__Board(const __Board& board) except +
		void set(const string& sfen)
		bool set_position(const string& position)
		bool set_hcp(char* hcp)
		bool set_psfen(char* psfen)
		void reset()
		string dump()
		void push(const int move)
		void pop()
		int peek()
		bool is_game_over()
		int isDraw(const int checkMaxPly)
		int move(const int from_square, const int to_square, const bool promotion)
		int drop_move(const int to_square, const int drop_piece_type)
		int move_from_usi(const string& usi)
		int move_from_csa(const string& csa)
		int move_from_move16(const unsigned short move16)
		int move_from_psv(const unsigned short move16)
		int turn()
		int ply()
		string toSFEN()
		string toCSAPos()
		void toHuffmanCodedPos(char* data)
		void toPackedSfen(char* data)
		int piece(const int sq)
		bool inCheck()
		int mateMoveIn1Ply()
		int mateMove(int ply);
		bool is_mate(int ply);
		unsigned long long getKey()
		bool moveIsPseudoLegal(const int move)
		bool moveIsLegal(const int move)
		vector[int] pieces_in_hand(const int color)
		vector[int] pieces()
		bool is_nyugyoku()
		void piece_planes(char* mem)
		void piece_planes_rotate(char* mem)
		void _dlshogi_make_input_features(char* mem1, char* mem2)
		void push_pass()
		void pop_pass()
		bool isOK()
		unsigned long long bookKey()

	int __piece_to_piece_type(const int p)

cdef class Board:
	cdef __Board __board

	def __cinit__(self, str sfen=None, Board board=None):
		cdef string sfen_b
		if sfen:
			sfen_b = sfen.encode('ascii')
			self.__board = __Board(sfen_b)
		elif board is not None:
			self.__board = __Board(board.__board)
		else:
			self.__board = __Board()

	def __copy__(self):
		return Board(board=self)

	def copy(self):
		return Board(board=self)

	def set_sfen(self, str sfen):
		cdef string sfen_b = sfen.encode('ascii')
		self.__board.set(sfen_b)

	def set_position(self, str position):
		cdef string position_b = position.encode('ascii')
		return self.__board.set_position(position_b)

	def set_hcp(self, np.ndarray hcp):
		return self.__board.set_hcp(hcp.data)

	def set_psfen(self, np.ndarray psfen):
		return self.__board.set_psfen(psfen.data)

	def reset(self):
		self.__board.reset()

	def __repr__(self):
		return self.__board.dump().decode('ascii')

	def push(self, int move):
		self.__board.push(move)

	def push_usi(self, str usi):
		cdef string usi_b = usi.encode('ascii')
		cdef int move = self.__board.move_from_usi(usi_b)
		if move != 0:
			self.__board.push(move)
		return move

	def push_csa(self, str csa):
		cdef string csa_b = csa.encode('ascii')
		cdef int move = self.__board.move_from_csa(csa_b)
		if move != 0:
			self.__board.push(move)
		return move

	def push_move16(self, unsigned short move16):
		cdef int move = self.__board.move_from_move16(move16)
		self.__board.push(move)
		return move

	def push_psv(self, unsigned short move16):
		cdef int move = self.__board.move_from_psv(move16)
		self.__board.push(move)
		return move

	def push_pass(self):
		assert not self.is_check()
		return self.__board.push_pass()

	def pop_pass(self):
		return self.__board.pop_pass()

	def pop(self):
		self.__board.pop()

	def peek(self):
		return self.__board.peek()

	def is_game_over(self):
		return self.__board.is_game_over()

	def is_draw(self, ply=None):
		cdef int _ply
		if ply:
			_ply = ply
		else:
			_ply = 2147483647
		return self.__board.isDraw(_ply)

	def move(self, int from_square, int to_square, bool promotion):
		return self.__board.move(from_square, to_square, promotion)

	def drop_move(self, int to_square, int drop_piece_type):
		return self.__board.drop_move(to_square, drop_piece_type)

	def move_from_usi(self, str usi):
		cdef string usi_b = usi.encode('ascii')
		return self.__board.move_from_usi(usi_b)

	def move_from_csa(self, str csa):
		cdef string csa_b = csa.encode('ascii')
		return self.__board.move_from_csa(csa_b)

	def move_from_move16(self, unsigned short move16):
		return self.__board.move_from_move16(move16)

	def move_from_psv(self, unsigned short move16):
		return self.__board.move_from_psv(move16)

	@property
	def legal_moves(self):
		return LegalMoveList(self)

	@property
	def turn(self):
		return self.__board.turn()

	@property
	def move_number(self):
		return self.__board.ply()

	def sfen(self):
		return self.__board.toSFEN().decode('ascii')

	def csa_pos(self):
		return self.__board.toCSAPos().decode('ascii')

	def to_hcp(self, np.ndarray hcp):
		return self.__board.toHuffmanCodedPos(hcp.data)

	def to_psfen(self, np.ndarray hcp):
		return self.__board.toPackedSfen(hcp.data)

	def piece(self, int sq):
		return self.__board.piece(sq)

	def piece_type(self, int sq):
		return 	__piece_to_piece_type(self.__board.piece(sq))

	def is_check(self):
		return self.__board.inCheck()

	def mate_move_in_1ply(self):
		return self.__board.mateMoveIn1Ply()

	def mate_move(self, int ply):
		assert ply % 2 == 1
		assert ply >= 3
		return self.__board.mateMove(ply)

	def is_mate(self, int ply):
		assert ply % 2 == 0
		return self.__board.inCheck() and self.__board.is_mate(ply)

	def zobrist_hash(self):
		return self.__board.getKey()

	def is_pseudo_legal(self, int move):
		return self.__board.moveIsPseudoLegal(move)

	def is_legal(self, int move):
		return self.__board.moveIsLegal(move)

	@property
	def pieces_in_hand(self):
		return (self.__board.pieces_in_hand(BLACK), self.__board.pieces_in_hand(WHITE))

	@property
	def pieces(self):
		return self.__board.pieces()

	def is_nyugyoku(self):
		return self.__board.is_nyugyoku()

	def piece_planes(self, np.ndarray features):
		return self.__board.piece_planes(features.data)

	def piece_planes_rotate(self, np.ndarray features):
		return self.__board.piece_planes_rotate(features.data)

	def _dlshogi_make_input_features(self, np.ndarray features1, np.ndarray features2):
		self.__board._dlshogi_make_input_features(features1.data, features2.data)

	def is_ok(self):
		return self.__board.isOK()

	def book_key(self):
		return self.__board.bookKey()

	def to_svg(self, lastmove=None, scale=1.0):
		import xml.etree.ElementTree as ET

		width = 230
		height = 192

		svg = ET.Element("svg", {
			"xmlns": "http://www.w3.org/2000/svg",
			"version": "1.1",
			"xmlns:xlink": "http://www.w3.org/1999/xlink",
			"width": str(width * scale),
			"height": str(height * scale),
			"viewBox": "0 0 {} {}".format(width, height),
		})

		defs = ET.SubElement(svg, "defs")
		for piece_def in SVG_PIECE_DEFS:
			defs.append(ET.fromstring(piece_def))

		if lastmove is not None:
			i, j = divmod(move_to(lastmove), 9)
			ET.SubElement(svg, "rect", {
				"x": str(20.5 + (8 - i) * 20),
				"y": str(10.5 + j * 20),
				"width": str(20),
				"height": str(20),
				"fill": "#f6b94d"
			})
			if not move_is_drop(lastmove):
				i, j = divmod(move_from(lastmove), 9)
				ET.SubElement(svg, "rect", {
					"x": str(20.5 + (8 - i) * 20),
					"y": str(10.5 + j * 20),
					"width": str(20),
					"height": str(20),
					"fill": "#fdf0e3"
				})

		svg.append(ET.fromstring(SVG_SQUARES))
		svg.append(ET.fromstring(SVG_COORDINATES))

		for sq in SQUARES:
			pc = self.__board.piece(sq)
			if pc != NONE:
				i, j = divmod(sq, 9)
				x = 20.5 + (8 - i) * 20
				y = 10.5 + j * 20

				ET.SubElement(svg, "use", {
					"xlink:href": "#{}".format(SVG_PIECE_DEF_IDS[pc]),
					"x": str(x),
					"y": str(y),
				})

		hand_pieces = [[], []]
		for c in COLORS:
			i = 0
			for hp, n in zip(HAND_PIECES, self.__board.pieces_in_hand(c)):
				if n >= 11:
					hand_pieces[c].append((i, NUMBER_JAPANESE_KANJI_SYMBOLS[n % 10]))
					i += 1
					hand_pieces[c].append((i, NUMBER_JAPANESE_KANJI_SYMBOLS[10]))
					i += 1
				elif n >= 2:
					hand_pieces[c].append((i, NUMBER_JAPANESE_KANJI_SYMBOLS[n]))
					i += 1
				if n >= 1:
					hand_pieces[c].append((i, HAND_PIECE_JAPANESE_SYMBOLS[hp]))
					i += 1
			i += 1
			hand_pieces[c].append((i, "手"))
			i += 1
			hand_pieces[c].append((i, "先" if c == BLACK else "後"))
			i += 1
			hand_pieces[c].append(( i, "☗" if c == BLACK else "☖"))

		for c in COLORS:
			if c == BLACK:
				x = 214
				y = 190
			else:
				x = -16
				y = -10
			scale = 1
			if len(hand_pieces[c]) + 1 > 13:
				scale = 13.0 / (len(hand_pieces[c]) + 1)
			for i, text in hand_pieces[c]:
				e = ET.SubElement(svg, "text", {
					"font-family": "serif",
					"font-size": str(14 * scale),
				})
				e.set("x", str(x))
				e.set("y", str(y - 14 * scale * i))
				if c == WHITE:
					e.set("transform", "rotate(180)")
				e.text = text

		return SvgWrapper(ET.tostring(svg).decode("utf-8"))
	
	def _repr_svg_(self):
		cdef int move = self.__board.peek()
		if move == 0:
			return self.to_svg()
		else:
			return self.to_svg(move)

def piece_to_piece_type(int p):
	return __piece_to_piece_type(p)

cdef extern from "cshogi.h":
	cdef cppclass __LegalMoveList:
		__LegalMoveList() except +
		__LegalMoveList(const __Board& board) except +
		bool end()
		int move()
		void next()
		int size()

	int __move_to(const int move)
	int __move_from(const int move)
	int __move_cap(const int move)
	bool __move_is_promotion(const int move)
	bool __move_is_drop(const int move)
	int __move_from_piece_type(const int move)
	int __move_drop_hand_piece(const int move)
	unsigned short __move16(const int move)
	unsigned short __move16_from_psv(const unsigned short move16)
	unsigned short __move16_to_psv(const unsigned short move16)
	int __move_rotate(const int move)
	string __move_to_usi(const int move)
	string __move_to_csa(const int move)
	int __move_from_csa(const string& csa)
	int __dlshogi_get_features1_num()
	int __dlshogi_get_features2_num()
	int __dlshogi_make_move_label(const int move, const int color)

cdef class LegalMoveList:
	cdef __LegalMoveList __ml

	def __cinit__(self, Board board):
		self.__ml = __LegalMoveList(board.__board)

	def __iter__(self):
		return self

	def __next__(self):
		if self.__ml.end():
			raise StopIteration()
		move = self.__ml.move()
		self.__ml.next()
		return move

	def __len__(self):
		return self.__ml.size()

def move_to(int move):
	return __move_to(move)

def move_from(int move):
	return __move_from(move)

def move_cap(int move):
	return __move_cap(move)

def move_is_promotion(int move):
	return __move_is_promotion(move)

def move_is_drop(int move):
	return __move_is_drop(move)

def move_from_piece_type(int move):
	return __move_from_piece_type(move)

def move_drop_hand_piece(int move):
	return __move_drop_hand_piece(move)

def move16(int move):
	return __move16(move)

def move16_from_psv(unsigned short move16):
	return __move16_from_psv(move16)

def move16_to_psv(unsigned short move16):
	return __move16_to_psv(move16)

def move_rotate(int move):
	return __move_rotate(move)

def move_to_usi(int move):
	return __move_to_usi(move).decode('ascii')

def move_to_csa(int move):
	return __move_to_csa(move).decode('ascii')

def move_from_csa(str csa):
	cdef string csa_b = csa.encode('ascii')
	return __move_from_csa(csa_b)

def opponent(int color):
	return BLACK if color == WHITE else WHITE

_dlshogi_FEATURES1_NUM = __dlshogi_get_features1_num()
_dlshogi_FEATURES2_NUM = __dlshogi_get_features2_num()
def _dlshogi_make_move_label(int move, int color):
	return __dlshogi_make_move_label(move, color)

cdef extern from "parser.h" namespace "parser":
	cdef cppclass __Parser:
		__Parser() except +
		string version
		vector[string] informations
		string sfen
		string endgame
		vector[string] names
		vector[float] ratings
		vector[int] moves
		vector[int] times
		vector[int] scores
		vector[string] comments
		string comment
		int win
		void parse_csa_file(const string& path) except +
		void parse_csa_str(const string& csa_str) except +

cdef class Parser:
	@staticmethod
	def parse_file(file, encoding=None):
		if type(file) is str:
			with open(file, 'r', encoding=encoding) as f:
				return Parser.parse_str(f.read())
		else:
			return Parser.parse_str(file.read())

	@staticmethod
	def parse_str(csa_str):
		parsers = []
		# split multiple matches
		matches = csa_str.split('\n/\n')
		for one_csa_str in matches:
			parser = Parser()
			parser.parse_csa_str(one_csa_str)
			parsers.append(parser)
		return parsers

	cdef __Parser __parser

	def __cinit__(self):
		self.__parser = __Parser()

	def parse_csa_file(self, str path):
		cdef string path_b = path.encode(locale.getpreferredencoding())
		self.__parser.parse_csa_file(path_b)

	def parse_csa_str(self, str csa_str):
		cdef string csa_str_b = csa_str.encode('utf-8')
		self.__parser.parse_csa_str(csa_str_b)

	@property
	def version(self):
		return self.__parser.version.decode('ascii')

	@property
	def var_info(self):
		d = {}
		for information in self.__parser.informations:
			k, v = information.decode('utf-8').split(':', 1)
			d[k[1:]] = v
		return d

	@property
	def sfen(self):
		return self.__parser.sfen.decode('ascii')

	@property
	def endgame(self):
		return self.__parser.endgame.decode('ascii')

	@property
	def names(self):
		return [name.decode('utf-8') for name in self.__parser.names]

	@property
	def ratings(self):
		return self.__parser.ratings

	@property
	def moves(self):
		return self.__parser.moves

	@property
	def times(self):
		return self.__parser.times

	@property
	def scores(self):
		return self.__parser.scores

	@property
	def comments(self):
		return [comment.decode('ascii') for comment in self.__parser.comments]

	@property
	def win(self):
		return self.__parser.win

	@property
	def comment(self):
		return self.__parser.comment.decode('utf-8')

cdef extern from "dfpn.h":
	cdef cppclass __DfPn:
		__DfPn() except +
		__DfPn(const int max_depth, const unsigned int max_search_node, const int draw_ply) except +
		bool search(__Board& board)
		bool search_andnode(__Board& board)
		void stop(const bool stop)
		int get_move(__Board& board)
		void get_pv(__Board& board)
		vector[unsigned int] pv
		void set_draw_ply(const int draw_ply)
		void set_maxdepth(const int depth)
		void set_max_search_node(int max_search_node)
		unsigned int get_searched_node()

cdef class DfPn:
	cdef __DfPn __dfpn

	def __cinit__(self, depth=31, nodes=1048576, draw_ply=32767):
		self.__dfpn = __DfPn(depth, nodes, draw_ply)

	def search(self, Board board):
		return self.__dfpn.search(board.__board)

	def search_andnode(self, Board board):
		return self.__dfpn.search_andnode(board.__board)

	def stop(self, bool stop):
		self.__dfpn.stop(stop)

	def get_move(self, Board board):
		return self.__dfpn.get_move(board.__board)

	def get_pv(self, Board board):
		self.__dfpn.get_pv(board.__board)
		return self.__dfpn.pv

	def set_draw_ply(self, int draw_ply):
		self.__dfpn.set_draw_ply(draw_ply)

	def set_max_depth(self, int max_depth):
		self.__dfpn.set_maxdepth(max_depth)

	def set_max_search_node(self, int max_search_node):
		self.__dfpn.set_max_search_node(max_search_node)

	@property
	def searched_node(self):
		return self.__dfpn.get_searched_node()



import cshogi
from cshogi.gym_shogi.envs import ShogiEnv

class ShogiVecEnv:
	metadata = {'render.modes': ['sfen']}

	def __init__(self, int num_envs):
		self.num_envs = num_envs
		self.envs = [ShogiEnv() for _ in range(num_envs)]
		
	def reset(self):
		for env in self.envs:
			env.reset()

	def render(self, mode='sfen'):
		return [env.render(mode='sfen') for env in self.envs]

	def step(self, list moves):
		cdef int i
		cdef int move

		rewards = []
		dones = []
		is_draws = []

		for i, move in enumerate(moves):
			reward, done, is_draw = self.envs[i].step(move)
			rewards.append(reward)
			dones.append(done)
			is_draws.append(is_draw)
			if done:
				self.envs[i].reset()
		
		return rewards, dones, is_draws



import subprocess
import os.path
import locale

class Engine:
    def __init__(self, cmd, connect=True, debug=False):
        self.cmd = cmd
        self.debug = debug
        if connect:
            self.connect()
        else:
            self.proc = None
            self.name = None

    def connect(self, listener=None):
        if self.debug: listener = print
        cwd = os.path.dirname(self.cmd)
        self.proc = subprocess.Popen([self.cmd], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd if cwd != '' else None)

        cmd = 'usi'
        if listener:
            listener(cmd)
        self.proc.stdin.write(cmd.encode('ascii') + b'\n')
        self.proc.stdin.flush()

        while True:
            self.proc.stdout.flush()
            line = self.proc.stdout.readline()
            if line == '':
                raise EOFError()
            line = line.strip()
            if line[:7] == b'id name':
                self.name = line[8:].decode('ascii')
            elif line == b'usiok':
                break
        if listener:
            listener(self.name)

    def usi(self, listener=None):
        if self.debug: listener = print
        cmd = 'usi'
        if listener:
            listener(cmd)
        self.proc.stdin.write(cmd.encode('ascii') + b'\n')
        self.proc.stdin.flush()

        lines = []
        while True:
            self.proc.stdout.flush()
            line = self.proc.stdout.readline()
            if line == '':
                raise EOFError()
            line = line.strip().decode(locale.getpreferredencoding())
            if listener:
                listener(line)
            if line == 'usiok':
                break
            lines.append(line)
        return lines

    def setoption(self, name, value, listener=None):
        if self.debug: listener = print
        cmd = 'setoption name ' + name + ' value ' + str(value)
        if listener:
            listener(cmd)
        self.proc.stdin.write(cmd.encode(locale.getpreferredencoding()) + b'\n')
        self.proc.stdin.flush()

    def isready(self, listener=None):
        if self.debug: listener = print
        cmd = 'isready'
        if listener:
            listener(cmd)
        self.proc.stdin.write(cmd.encode('ascii') + b'\n')
        self.proc.stdin.flush()

        while True:
            self.proc.stdout.flush()
            line = self.proc.stdout.readline()
            if line == '':
                raise EOFError()
            line = line.strip().decode('shift-jis')
            if listener:
                listener(line)
            if line == 'readyok':
                break

    def usinewgame(self, listener=None):
        if self.debug: listener = print
        cmd = 'usinewgame'
        if listener:
            listener(cmd)
        self.proc.stdin.write(cmd.encode('ascii') + b'\n')
        self.proc.stdin.flush()

    def position(self, moves=None, sfen="startpos", listener=None):
        if self.debug: listener = print
        cmd = 'position ' + sfen
        if moves:
            cmd += ' moves ' + ' '.join(moves)
        if listener:
            listener(cmd)
        self.proc.stdin.write(cmd.encode('ascii') + b'\n')
        self.proc.stdin.flush()

    def go(self, ponder=False, btime=None, wtime=None, byoyomi=None, binc=None, winc=None, nodes=None, listener=None):
        if self.debug: listener = print
        cmd = 'go'
        if ponder:
            cmd += ' ponder'
        if btime is not None:
            cmd += ' btime ' + str(btime)
        if wtime is not None:
            cmd += ' wtime ' + str(wtime)
        if byoyomi is not None:
            cmd += ' byoyomi ' + str(byoyomi)
        else:
            if binc is not None:
                cmd += ' binc ' + str(binc)
            if winc is not None:
                cmd += ' winc ' + str(winc)
        if nodes is not None:
            cmd += ' nodes ' + str(nodes)
        if listener:
            listener(cmd)
        self.proc.stdin.write(cmd.encode('ascii') + b'\n')
        self.proc.stdin.flush()

        while True:
            self.proc.stdout.flush()
            line = self.proc.stdout.readline()
            if line == '':
                raise EOFError()
            line = line.strip().decode(locale.getpreferredencoding())
            if listener:
                listener(line)
            if line[:8] == 'bestmove':
                items = line[9:].split(' ')
                if len(items) == 3 and items[1] == 'ponder':
                    return items[0], items[2]
                else:
                    return items[0], None

    def go_mate(self, byoyomi=None, listener=None):
        if self.debug: listener = print
        cmd = 'go mate'
        if byoyomi is not None:
            cmd += ' ' + str(byoyomi)
        else:
            cmd += ' infinite'
        if listener:
            listener(cmd)
        self.proc.stdin.write(cmd.encode('ascii') + b'\n')
        self.proc.stdin.flush()
        while True:
            self.proc.stdout.flush()
            line = self.proc.stdout.readline()
            if line == '':
                raise EOFError()
            line = line.strip().decode(locale.getpreferredencoding())
            if listener:
                listener(line)
            if line[:9] == 'checkmate':
                items = line[10:]
                return items

    def ponderhit(self, listener=None):
        if self.debug: listener = print
        cmd = 'ponderhit'
        if listener:
            listener(cmd)
        self.proc.stdin.write(cmd.encode('ascii') + b'\n')
        self.proc.stdin.flush()

    def stop(self, listener=None):
        if self.debug: listener = print
        cmd = 'stop'
        if listener:
            listener(cmd)
        self.proc.stdin.write(cmd.encode('ascii') + b'\n')
        self.proc.stdin.flush()

    def gameover(self, result=None, listener=None):
        if self.debug: listener = print
        cmd = 'gameover'
        if result:
            cmd += ' ' + result
        if listener:
            listener(cmd)
        self.proc.stdin.write(cmd.encode('ascii') + b'\n')
        self.proc.stdin.flush()

    def quit(self, listener=None):
        if self.debug: listener = print
        cmd = 'quit'
        if listener:
            listener(cmd)
        self.proc.stdin.write(cmd.encode('ascii') + b'\n')
        try:
            self.proc.stdin.flush()
        except BrokenPipeError:
            pass
        self.proc.wait()
        self.proc = None
